;**********************************************************************
;    Filename: usbpsu.asm                                             *
;    Description:  USB mini power supply control firmware             *
;    Author: H.Momose (hirofumi@momose.com)                           *
;    Copyright(c) 2007 H.Momose   All rights reserved                 *
;    Includes software serial program by picfun.com                   *
;**********************************************************************
;    Files required: p16F819.inc                                      *
;**********************************************************************
;    Revised history                                                  *
;    Ver  Date        Remarks                                         *
;    1.00 2007/07/24  First release                                   *
;    1.10 2007/08/05  Add firmware updater                            *
;                     Fix problem device does not reterun from sleep, *
;                     when operates CC just before go sleep.          *
;                     Adjust serial timing more precisely.            *
;                                                                     *
;**********************************************************************

	list      p=16f819            ; list directive to define processor
	#include <p16F819.inc>        ; processor specific variable definitions

	errorlevel  -302              ; suppress message 302 from list file

	
	__CONFIG   _CP_OFF & _WRT_ENABLE_OFF & _CPD_OFF & _CCP1_RB2 & _DEBUG_OFF & _LVP_OFF & _BODEN_OFF & _MCLR_OFF & _WDT_OFF & _PWRTE_ON & _HS_OSC 
; '__CONFIG' directive is used to embed configuration word within .asm file.
; The lables following the directive are located in the respective .inc file.
; See data sheet for additional information on configuration word settings.

; PORTA Bits
VIN			equ				H'00'		;(A/17) Voltage measurement in
AIN			equ				H'01'		;(A/18) Current measurement in
VREFN		equ				H'02'		;(A/01) V reference -
VREFP		equ				H'03'		;(A/02) V reference +
RXD			equ				H'04'		;(I/03) Serial data in
ADPTR		equ				H'05'		;(I/04) AC adaptor detection (Input only)
XOUT			equ				H'06'		;(X/15) Xtal
XIN			equ				H'07'		;(X/16) Xtal

; PORTB Bits
PWREN		equ				H'00'		;(I/06) Wakeup (Negative)
USBPWR		equ				H'01'		;(O/07) USB buspower to main circuit (Negative)
PWM			equ				H'02'		;(O/08) PWM out
OUTSW		equ				H'03'		;(I/09) Output SW
TXD			equ				H'04'		;(O/10) Serial data out
VCCOK		equ				H'05'		;(I/11) VCC detection
CVLED		equ				H'06'		;(O/12) CV LED Green
CCLED		equ				H'07'		;(O/13) CC LED Red

#define CVLED_MASK				B'01000000'
#define CCLED_MASK				B'10000000'

; GRPs
vset_h		equ				0x20			; Voltage set (H)
vset_l		equ				0x21			; Voltage set (L)
aset_h		equ				0x22			; Current set (H)
aset_l		equ				0x23			; Current set (L)

vread_h		equ				0x24			; Voltage measured (H)
vread_l		equ				0x25			; Voltage measured (L)
aread_h		equ				0x26			; Current measured (H)
aread_l		equ				0x27			; Current measured (L)

vdiff_h		equ				0x28			; Voltage difference (H)
vdiff_l		equ				0x29			; Voltage difference (L)
adiff_h		equ				0x2a			; Current difference (H)
adiff_l		equ				0x2b			; Current difference (L)
vdiff_pre1_h	equ				0x2c			; Prev voltage difference (H)
vdiff_pre1_l	equ				0x2d			; Prev voltage difference (L)
adiff_pre1_h	equ				0x2e			; Prev current difference (H)
adiff_pre1_l	equ				0x2f			; Prev current difference (L)

dpwm_h		equ				0x30			; PWM delta (H)
dpwm_l		equ				0x31			; PWM delta (L)

maxpwm_h		equ				0x32			; PWM maximum (H)
maxpwm_l		equ				0x33			; PWM maximum (L)

advread_h		equ				0x34			; A/D voltage result (H) 
advread_l		equ				0x35			; A/D voltage result (L)
adaread_h		equ				0x36			; A/D current result (H)
adaread_l		equ				0x37			; A/D current result (L)

adstat		equ				0x38			; A/D routine state (one of followings)
#define ADSTAT_SAMPLE_VOLTAGE	1			;  Voltage sampling in progress
#define ADSTAT_CONVERT_VOLTAGE	2			;  Voltage conversion in progress
#define ADSTAT_SAMPLE_CURRENT	3			;  Current sampling in progress
#define ADSTAT_CONVERT_CURRENT	4			;  Current conversion in progress
#define ADSTAT_DONE			7			;  All A/D operation has done

psustat		equ				0x39			; PSU status flag (any bit of followings)
#define F_OUTPUT				0			;  Output sw
#define F_CCMODE				1			;  CC mode
#define F_ADAPTR				2			;  Adaptor connected
#define F_VCCOK				3			;  Adaptor power OK

#define OUTPUT_MASK			B'00000001'

tmp_l		equ				0x3a			; tmp
tmp_h		equ				0x3b

_AL			equ				0x3c			; used for 16bit math functions
_AH			equ				0x3d
_BL			equ				0x3e
_BH			equ				0x3f

txleft		equ				0x40			; TX left bytes
rxleft		equ				0x41			; RX left bytes
;pbuff		equ				0x42			; TX/RX data buffer pointer

comstat		equ				0x43			; Serial state
#define RXWAITING				0			;  Waiting for RX data

outputsw_cnt	equ				0x44			; Output sw stable counter

dbg_h		equ				0x45			; For debug use
dbg_l		equ				0x46


buff_0		equ				0x50			; Data buffer (0x50 - 0x5f)
buff_1		equ				0x51
buff_2		equ				0x52
buff_3		equ				0x53
buff_4		equ				0x54
buff_5		equ				0x55
buff_6		equ				0x56
buff_7		equ				0x57
buff_8		equ				0x58
buff_9		equ				0x59
buff_10		equ				0x5a
buff_11		equ				0x5b
buff_12		equ				0x5c
buff_13		equ				0x5d
buff_14		equ				0x5e
buff_15		equ				0x5f

#define		VERSION_L1		"1"			; Firm version string (4 letter)
#define		VERSION_L2		"."
#define		VERSION_L3		"1"
#define		VERSION_L4		"0"

#define 		TSAMPLE_H			0xff			; A/D sampling time
#define		TSAMPLE_L			0x9c			; (0xffff - 0xffb5 +1) * 0.2us (=20us) + ISR latency

#define		MAXVSET_H			0x03			; Maximum voltage setting
#define		MAXVSET_L			0xf0
#define		MAXASET_H			0x03			; Maximum current setting
#define		MAXASET_L			0xf0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Following code from picfun.com
; http://www.picfun.com/serialframe.html

#define		T0IF		TMR0IF
#define		T0IE		TMR0IE

#DEFINE		TXBUSY	RESULT,0    ;MrW[
#DEFINE		RXBUSY	RESULT,2    ;MrW[
#DEFINE		RECV		RESULT,3    ;M
#DEFINE		RXERR	RESULT,4    ;MG[tO

;****  Define Variables for RS-232C Comm ***** 
RESULT		EQU     0x70     ;MXe[^X
WREG			EQU     0x71     ;wreg save area
SREG			EQU     0x72     ;status reg save area
BITCNT		EQU     0x73     ;rbgJE^
EXTRA		EQU     0x74     ;Xgbvrbg
TXREG		EQU     0x75     ;Mf[^obt@
RXREG		EQU     0x76     ;Mf[^obt@
;CNT2		EQU     0x67     ;XL[px
BITTMR1 		SET     0x77     ;XLbvprbg

;********  ʐMx̐ݒƃp[^ݒ ****
; ʐMx̓^C}0̊ݎƊ݂̉񐔃JE^Ō܂B
; ʐMx  pX Tcy prescaler BITTMR@BAUDRATE
;   300    3333.33  8333    64    @130(7E)  @5
;  1200    833.33@ 2083    16    @130(7E)  @3
;  2400    416.67   1041     8  @@128(80)  @2
;  4800    208.33    521     4    @124(84)   @1
;  9600    104.17    260     2    @122(86)  @0
;
;@BAUDRATEʐMxw̃p[^ƂLlŎw肷B
;@@iNbNPOMH̏ꍇj
;@@ 5=300bps  3=1200bps  2=2400bps  1=4800bps  0=9600bps
;
;@L9600bps̏ꍇ̗
;BAUDRATE    SET 0   ;vXP[l
;BITTMR      SET 086H    ;TMR0 DATA 134

BAUDRATE    SET B'10000001'		;vXP[l(OPTION REGl) 0x80/RBPU
BITTMR      SET 0x83			;TMR0 DATA 131

; ========================================================================================
; Vector
; Input  : None
; Output : None
; Temp   : None
; Remks  : 
; ========================================================================================
	ORG		0x000			; Reset vector
	goto	updater			; goto updater

	ORG		0x004			; Interrupt vector
	; Save context
	MOVWF	WREG				; Save w reg
	SWAPF	STATUS,W			; Status to wreg
	CLRF	STATUS
	MOVWF	SREG  	      	; Save status
	banksel	0				; Ensure bank 0

	; Reason of interrupt
	;  Serial TMR0 => INTCON:TMR0IF 
	;  A/D sampling done TMR1 => PIR1:TMR1IF
	;  A/D completion => PIR1:ADIF

	; Check if serial TMR0 interrupts
	btfss	INTCON, TMR0IE
	goto	isr_main
	btfsc	INTCON, TMR0IF
	goto	INT232
	goto	isr_main

; ========================================================================================
; LED blinker update mode notifyer
; Input reg    : 
; Output reg   : 
; Internal reg : vset_l, vset_h
; Remarks      : Tell user that the device is in updater mode.
; ========================================================================================
blinker
	incf	vset_l, F
	btfss	STATUS, Z
	return
	incf	vset_h, F
	btfss	STATUS, Z
	return
	movfw	PORTB
	xorlw	CVLED_MASK
	movwf	PORTB
	return

; ========================================================================================
; Onboard flash program updater
; Input reg    : 
; Output reg   : w
; Internal reg : tmp_l, tmp_h, vset_l, txleft, rxleft, comstat, buff_0 - buff_15
; Remarks      : Require RS-232C comm module
; ========================================================================================
updater
	; Check OUTPUT sw to determine which program to start
	clrf	tmp_l
	clrf	tmp_h
updater_pre_stable_wait
	incf	tmp_l, F
	btfss	STATUS, Z
	goto	updater_pre_stable_wait
	incf	tmp_h, F
	btfss	STATUS, Z
	goto	updater_pre_stable_wait

	clrf	tmp_h
	movlw	0x10
	movwf	vset_l
updater_check_sw
	incf	tmp_l, F
	btfss	STATUS, Z
	goto	updater_check_sw
	
	btfsc	PORTB, OUTSW			; If this code is placed just before tmp_l loop, it is too release-sensitive. 
	goto	main					; SW relaesed -> start main program.

	incf	tmp_h, F
	btfss	STATUS, Z
	goto	updater_check_sw
	decf	vset_l, F
	btfss	STATUS, Z
	goto	updater_check_sw

updater_main
	; Port direction
	banksel	TRISA
	movlw	B'11111111'			; Setup PORTA XI/XO/ADPTR/RXD/VREFP/VREFN/AIN/VIN
	movwf	TRISA
	movlw	B'00101001'			; Setup PORTB CCLED/CVLED/VCCOK/TXD/OUTSW/PWM/USBPWR/PWREN
	movwf	TRISB
	movlw	B'10001101'			; Configure A/D (See D.S. P82)
	movwf	ADCON1

	; Clear ALL I/O ports
	banksel	PORTA
	clrf	PORTA
	clrf	PORTB

	; Clear working regs
	clrf	txleft
	clrf	rxleft
	clrf	comstat

	; Initialize serial module
	call	OPEN232

	bsf		INTCON, GIE

updater_loop
	call	blinker				; CV LED blinker

	btfsc	TXBUSY				; Do nothind if TX is in progress
	goto	updater_loop

	movf	txleft, F				; Do tx if there are data to be sent
	btfsc	STATUS, Z
	goto	updater_rx
	
	; Send data from buffer
	movfw	INDF					; TXREG = *buffer
	movwf	TXREG

	incf	FSR, F				; buffer++
	decf	txleft, F				; txleft--

	btfsc	STATUS, Z				; Prepare RX, if no more data
	bcf		comstat, RXWAITING 

	call	PUTCHAR
	goto	updater_loop

updater_rx
	; Recieve
	btfss	comstat, RXWAITING		; if (!RXWAITING) call GETCHAR
	call	GETCHAR
	bsf		comstat, RXWAITING

	btfsc	RXERR				; if (RXERR) goto updater_rxerr
	goto	updater_rxerr

	btfss	RECV					; Check data arrival
	goto	updater_loop			; No data

	movf	rxleft, F				; if (rxleft != 0) goto comm_getoperand
	btfss	STATUS, Z
	goto	updater_getoperand

	; Updater command dispatcher
	movlw	"I"					; Ident updater
	xorwf	RXREG, W
	btfsc	STATUS, Z
	goto	updater_ident

	movlw	"E"					; Erase flash memory
	xorwf	RXREG, W
	btfsc	STATUS, Z
	goto	updater_erase

	movlw	"W"					; Write data to flash memory
	xorwf	RXREG, W
	btfsc	STATUS, Z
	goto	updater_write

	movlw	"Q"					; Quit updater and start main program
	xorwf	RXREG, W
	btfsc	STATUS, Z
	goto	updater_quit

	goto	updater_ack_ng

updater_rxerr
	call	GETCHAR
	goto	updater_loop

updater_getoperand
	movfw	RXREG
	call	hex2int				; Convert hex character to integer

	btfsc	rxleft, 0				; Branch off even/odd so that 2 bytes hex char makes 1 byte binary 
	goto	updater_getoperand_odd

updater_getoperand_even
	; DATA is MS 4 bit
	movwf	INDF
	swapf	INDF, F
	goto	updater_getoperand_next

updater_getoperand_odd
	; DATA is LS 4 bit
	addwf	INDF, F
	incf	FSR, F

updater_getoperand_next
	decf	rxleft, F				; rxleft --
	
	btfss	STATUS, Z				; if (rxleft > 0) goto updater_getoperand_more
	goto	updater_getoperand_more
	
	; Now 8 bytes (4 words = 1 block) of code is being stored in the buffer
	; Write this block to FLASH memory.
	bcf		INTCON, GIE
	bsf		PORTB, CCLED

	banksel	EECON1
	bsf		EECON1, EEPGD
	bsf		EECON1, WREN
	bcf		EECON1, FREE
	
	; Set buffer head to pointer
	movlw	buff_0
	movwf	FSR

updater_getoperand_write
	banksel	EEDATH
	movfw	INDF					; EEDATH(MSB) = *buff++
	movwf	EEDATH
	incf	FSR, F
	movfw	INDF					; EEDATA(LSB) = *buff++
	movwf	EEDATA
	incf	FSR, F

	; Perform req'd sequence (see D.S. P31)
	banksel	EECON1
	movlw	0x55
	movwf	EECON2
	movlw	0xaa
	movwf	EECON2
	bsf		EECON1, WR
	nop
	nop
	
	; Increment address
	banksel	EEADR
	incf	EEADR, F
	btfsc	STATUS, Z
	incf	EEADRH, F
	btfsc	STATUS, Z				; if(EEADRH == 0) goto updater_getoperand_end
	goto	updater_getoperand_end	; This register has 3 effective bits on LSB 
	
	movfw	FSR					; if(buff != buff_8) goto updater_getoperand_write
	xorlw	buff_8
	btfss	STATUS, Z
	goto	updater_getoperand_write

	banksel	EECON1
	bcf		EECON1, WREN
	banksel	0
	bsf		INTCON, GIE
	bcf		PORTB, CCLED
	goto	updater_ack_ok

updater_getoperand_end
	movlw	0x02					; Be sure of EEADR never indicate updater resident address. 
	movwf	EEADRH				; Set 0x0200 for safety.
	movlw	0x00
	movwf	EEADR

	banksel	EECON1
	bcf		EECON1, WREN
	banksel	0
	bsf		INTCON, GIE
	bcf		PORTB, CCLED
	goto	updater_ack_ok

updater_getoperand_more
	call	GETCHAR
	goto	updater_loop

updater_ident
	; Send UPDATER<CRLF> , length=9
	movlw	"U"
	movwf	buff_0
	movlw	"P"
	movwf	buff_1
	movlw	"D"
	movwf	buff_2
	movlw	"A"
	movwf	buff_3
	movlw	"T"
	movwf	buff_4
	movlw	"E"
	movwf	buff_5
	movlw	"R"
	movwf	buff_6
	movlw	0x0d
	movwf	buff_7
	movlw	0x0a
	movwf	buff_8
	movlw	9
	movwf	txleft
	movlw	buff_0
	movwf	FSR
	goto	updater_loop

updater_erase
	; FLASH erase
	bcf		INTCON, GIE
	bsf		PORTB, CCLED

	; Perform flash erase
	; Load start addr (0x02:00)
	banksel	EEADRH
	movlw	0x02					; Hi addr
	movwf	EEADRH
	movlw	0x00					; Lo addr (must be multiple of 32)
	movwf	EEADR
	
updater_erase_row
	banksel	EECON1
	bsf		EECON1, EEPGD
	bsf		EECON1, WREN
	bsf		EECON1, FREE
	
	; Perform req'd sequence (D.S. P29)
	movlw	0x55
	movwf	EECON2
	movlw	0xaa
	movwf	EECON2
	bsf		EECON1, WR
	nop
	nop
	bcf		EECON1, FREE
	bcf		EECON1, WREN
	
	banksel	EEADR
	movlw	0x20
	addwf	EEADR, F
	btfsc	STATUS, C
	incf	EEADRH, F
	movf	EEADRH, F				; Only 3 LS bits are effective on EEADRH (MS 5 bits are not implemented, always read as 0) 
	btfss	STATUS, Z
	goto	updater_erase_row
	; Erase END

	; Set start address to prepare subsequent write operation	
	movlw	0x02 				; Hi addr
	movwf	EEADRH
	movlw	0x00					; Lo addr
	movwf	EEADR

	banksel	0
	bsf		INTCON, GIE
	bcf		PORTB, CCLED
	goto	updater_ack_ok
	
updater_write
	; Recive data to be written
	movlw	0x10					; 16 bytes to recieve (8byte binary)
	movwf	rxleft
	movlw	buff_0				; Set buffer head pointer
	movwf	FSR
	call	GETCHAR
	goto	updater_loop
	
updater_quit
	; Quit updater and start main program.
	goto	main
	
updater_ack_ok
	; Send OK<CRLF>, length=4
	movlw	"O"
	movwf	buff_0
	movlw	"K"
	movwf	buff_1
	movlw	0x0d
	movwf	buff_2
	movlw	0x0a
	movwf	buff_3
	movlw	4
	movwf	txleft
	movlw	buff_0
	movwf	FSR
	goto	updater_loop

updater_ack_ng
	; Send NG<CRLF>, length=4
	movlw	"N"
	movwf	buff_0
	movlw	"G"
	movwf	buff_1
	movlw	0x0d
	movwf	buff_2
	movlw	0x0a
	movwf	buff_3
	movlw	4
	movwf	txleft
	movlw	buff_0
	movwf	FSR
	goto	updater_loop

; ========================================================================================
; Convert HEX character to integer (w <= toINT(w))
; Input reg    : w
; Output reg   : w
; Internal reg : tmp_l
; Remarks      : 
; ========================================================================================
hex2int
	movwf	tmp_l
	sublw	0x2f					; 0x2f - w, if C=1 out of range
	btfsc	STATUS, C
	goto	hex2int_notdigit
	movfw	tmp_l
	sublw	0x39					; 0x39 - q, if C=0 out of range
	btfss	STATUS, C
	goto	hex2int_notdigit
	movlw	"0"
	subwf	tmp_l, W
	return

hex2int_notdigit
	movfw	tmp_l
	sublw	0x40					; 0x40 - w, if C=1 out of range
	btfsc	STATUS, C
	retlw	0x00
	movfw	tmp_l
	sublw	0x46					; 0x3a - q, if C=0 out of range
	btfss	STATUS, C
	retlw	0x00
	movlw	"A"
	subwf	tmp_l, W
	addlw	0x0a
	return

; ========================================================================================
; RS-232C software serial communication program
; From http://www.picfun.com/docs/rs232cv2.asm
; TO SEND
;   Set data to TXREG then call PUTCHAR, TXBUSY will go down when transmission has done.
; TO RECV
;   Call GETCHAR, data will be available when RECV go up without ERROR
; CONNECTION
;   RXD (PIN3 RA4) Pin constrained because this function uses hardware timer interruption.
;   TXD (PIN4 RB4) Can use any pin. Placed by PCB layout.

;*************************************
;  vZX
;*************************************
OPEN232
	CLRF	RESULT		;Xe[^X̃NA
	BSF		STATUS,RP0	;Set page 1
	BCF		TRISB,4		;Set TXD (RB4) 
	BSF		TRISA,4		;Set RXD (RA4)
	BCF		STATUS,RP0	;Set Page 0
	;BSF		PORTA,5		;TXD SET TO H
	BSF		PORTB,4
	;BSF		INTCON,GIE
	RETURN

;************************************
; PoCgMvZX
;@Mf[^TXREGɃZbgCALL
;************************************
PUTCHAR
	BSF		TXBUSY		;MrW[Zbg
	MOVLW	8			;rbgJE^l
	MOVWF	BITCNT		;lZbg
	MOVLW	1			;Xgbvrbg=1
	MOVWF	EXTRA		;XgbvrbgZbg
	CALL	TXSTART		;X^[grbgMJn
	BCF		INTCON,T0IF	;݃tOZbg
	BSF		INTCON,T0IE 	;^C}0݋
	RETURN
;**** X^[grbgM ****
TXSTART
	CLRF	TMR0			;^C}Zbg
	BSF		STATUS,RP0	;page 1
	;CLRWDT				;don't care
	MOVLW	BAUDRATE		;set prescaler
	MOVWF	OPTION_REG
	BCF		STATUS,RP0	;page 0 
	;BCF		PORTA,5		;X^[grbgo
	;BCF		PORTB,3
	bcf		PORTB, TXD
	MOVLW	BITTMR		;rbg^C}Zbg
	MOVWF	TMR0			;start timer
	RETURN

;*************************************
;  PoCgMvZX
;*************************************
GETCHAR
	BCF		RECV			;MtONA
	MOVLW	9			;rbgJE^l
	MOVWF	BITCNT		;lZbg
	CLRF	RXREG		;obt@NA
	BCF		RXERR		;G[tONA
;****  X^[grbgM҂  ****
	CLRF	TMR0			;^C}Zbg
	BSF		STATUS,RP0	;page 1
	;CLRWDT				;dont't care
	;MOVLW	038H
	MOVLW	0B8H			; with /RBPU
	MOVWF	OPTION_REG
	BCF		STATUS,RP0	;page 0
	MOVLW	0FFH			;tJEglZbg
	MOVWF	TMR0			;
	BCF		INTCON,T0IF	;^C}0݃tOZbg
	BSF		INTCON,T0IE	;^C}0݋
	RETURN				;ݑ҂

;****************************************
; Interrupt process module
;****************************************
INT232
	;BTFSS	INTCON,T0IF	;TMR0's interrupt?
	;RETFIE				;not, quick return
	BCF		INTCON,T0IF	;reset T0IF bit
	
;****  save W register & status register
	;MOVWF	WREG			;save w reg
	;SWAPF	STATUS,W		;status to wreg
	;CLRF	STATUS
	;MOVWF	SREG        	;save status

	BTFSC	TXBUSY		;MH
	GOTO	TXNEXT		;Mݏ
	BTFSC	RXBUSY		;MH
	GOTO	RXNEXT		;Mݏ
	GOTO	STARTBIT		;X^[grbgM
;****  register restore and return
RESTORE
	SWAPF	SREG,W		;get saved status
	MOVWF	STATUS
	SWAPF	WREG,F		;get saved wreg
	SWAPF	WREG,W
	RETFIE

;***************************************
;  MݏvZX
;***************************************
TXNEXT
	MOVLW	BITTMR		;^C}OăX^[g
	MOVWF	TMR0
	MOVF	BITCNT,W		;rbgJEgmF
	BTFSC	STATUS,Z		;rbgI
	GOTO	STOPBIT		;XgbvrbgM
NEXTTXBIT
	DECF	BITCNT,F		;rbgJE^Z
	BSF		STATUS,C		;C on
	RRF		TXREG,F		;Mf[^Vtg
	BTFSS	STATUS,C		;OP
	bcf		PORTB, TXD
	BTFSC	STATUS,C
	bsf		PORTB, TXD
	GOTO	RESTORE		;return
;****  XgbvrbgMvZX  ****
STOPBIT
	MOVF	EXTRA,W		;XgbvrbgM
	BTFSC	STATUS,Z		;end test
	GOTO	DONE			;SrbgM
	DECF	EXTRA,F		;XgbvrbgZ
	bsf		PORTB, TXD
	GOTO	RESTORE		;߂
;***  all end of send ****
DONE
	;bsf		STATUS, RP0	;page 1
	;MOVLW	0B8H
	;MOVWF	OPTION_REG
	;bcf		STATUS, RP0	;page 0
	;bcf		INTCON, T0IF
	BCF		INTCON,T0IE	;^C}O݋֎~
	BCF		TXBUSY		;MrW[tONA
	GOTO	RESTORE		;߂

;*****************************************
;@X^[grbgMݏvZX
;*****************************************
STARTBIT
	; CLRWDT				;WDTZbg
	; Don't care noise due to onboard USB-Serial converter. Save CPU cycles.
	;CALL	TIME10		;10usecx
	;BTFSC	PORTA,4		;ēxX^[grbgmF
	;GOTO	NOISE		;G[^[
	BSF		RXBUSY		;MrW[Zbg
	CLRF	TMR0			;^C}
	BSF		STATUS,RP0	;page 1
	MOVLW	BAUDRATE		;rbgZbg
	MOVWF	OPTION_REG
	BCF		STATUS,RP0	;page 0
	MOVLW	BITTMR1		;rbg1.1{Zbg
	MOVWF	TMR0			;X^[grbgXLbv
	GOTO 	RESTORE		;߂
;*** X^[grbgG[iȂj ****
;NOISE       
;	MOVLW	0FFH		;ēxX^[grbgM҂
;	MOVWF	TMR0
;	GOTO	RESTORE		;߂

;*****************************************
;@MݏvZX
;*****************************************
RXNEXT
	BSF		STATUS,RP0	;page 1
	;CLRWDT				;^C}
	MOVLW	BAUDRATE
	MOVWF	OPTION_REG
	BCF		STATUS,RP0	;page 0
	MOVLW	BITTMR
	MOVWF	TMR0			;ăX^[g
	DECFSZ	BITCNT,F		;rbgJEgIH
	GOTO	NEXTRXBIT
;*** XgbvrbgM  ***
	BTFSS	PORTA,4		;Xgbvrbg̊mF
	BSF		RXERR		;G[tOON
	BCF		INTCON,T0IE	;݋֎~
	BCF		RXBUSY		;MrW[tONA
	BSF		RECV			;Mf[^LtOZbg
	GOTO	RESTORE

;**** c̃rbgM  ***
NEXTRXBIT
	BCF		STATUS,C		;rbgOZbg
	BTFSC	PORTA,4		;MrbgOP
	BSF		STATUS,C		;1 ̎PɃZbg
	RRF		RXREG,F		;RXREGɂPrbg
	GOTO	RESTORE		;߂

;*****  10usec x^C}
;TIME10                  ;10usec
;	MOVLW	15
;	MOVWF	CNT2        
;T_LP2
;	DECFSZ	CNT2,F		;2+3*15-1=46     
;	GOTO	T_LP2       
;	RETURN				;46+2=48

;********  End of RS-232C software serial communication program ***********
;**************************************************************************


; ========================================================================================
; Area from 0x0200 to 0x7fff is a target of updater. 
; All except updater must be placed on this area.
; ========================================================================================
	ORG		0x200

; ========================================================================================
; USBPSU main control program
; Input reg    : 
; Output reg   : 
; Internal reg : 
; Remarks      : 
; ========================================================================================
main
	; Port direction
	banksel	TRISA
	movlw	B'11111111'			; Setup PORTA XI/XO/ADPTR/RXD/VREFP/VREFN/AIN/VIN
	movwf	TRISA
	movlw	B'00101001'			; Setup PORTB CCLED/CVLED/VCCOK/TXD/OUTSW/PWM/USBPWR/PWREN
	movwf	TRISB
	movlw	B'10001101'			; Configure A/D (See D.S. P82)
	movwf	ADCON1

	; Clear ALL I/O ports
	banksel	PORTA
	clrf	PORTA
	clrf	PORTB

	; Clear working regs
	clrf	txleft
	clrf	rxleft
	clrf	comstat
	clrf	vset_h
	clrf	vset_l
	clrf	aset_h
	clrf	aset_l
	clrf	psustat
	clrf	vdiff_pre1_h
	clrf	vdiff_pre1_l
	clrf	adiff_pre1_h
	clrf	adiff_pre1_l
	clrf	dbg_h
	clrf	dbg_l
	clrf	adstat

	; Initialize serial module
	call	OPEN232

	movlw	0x07					; All jump tables are located above 0x700
	movwf	PCLATH				; Set PCLATH in advance

	; Setup TMR1
	movlw	B'00000100'			; Prescaler=1/1, osc=off, no T1 sync, int clk, tmr off
	movwf	T1CON

	; PWM ON
	movlw	B'00001100'
	movwf	CCP1CON	
	bsf		T2CON, TMR2ON

	; Setup Interrupt
	banksel	PIE1
	bsf		PIE1, TMR1IE			; Enable TMR1 INT
	bsf		PIE1, ADIE			; Enable A/D INT
	banksel	INTCON
	bsf		INTCON, PEIE			; Enable peripheral INT
	bsf		INTCON, GIE			; Enable Global INT

; ========================================================================================
; Loop head
loop

; ========================================================================================
; Sleep control
; When PORTB:PWREN is Low -> Operate / High -> Sleep
	btfss	PORTB, PWREN
	goto	check_sleep_end
	
	; Prepare for sleep
	bcf		INTCON, GIE			; Disable any INT

	bcf		psustat, F_OUTPUT		; Output flag off
	clrf	CCPR1L				; Reset PWM
	clrf	CCP1CON
	bcf		T2CON, TMR2ON

	movfw	PORTB				; Because power MOS gate is large capacitive. 
	movwf	tmp_l				; Avoid direct port configuration to ensure port level.
	bcf		tmp_l, PWM			; Main FET OFF
	bsf		tmp_l, USBPWR			; USB power FET OFF (its P channel)
	bcf		tmp_l, CVLED			; Turn off LEDs
	bcf		tmp_l, CCLED
	movfw	tmp_l
	movwf	PORTB
	
	; Prepare for wakeup
	banksel	OPTION_REG		
	bcf		OPTION_REG, INTEDG	; Set negative edge
	banksel	INTCON
	bcf		INTCON, INTF			; Clear INT flag
	bsf		INTCON, INTE			; Enable INT
	sleep						; Sleep...

	; Wakeup
	bcf		PORTB, USBPWR			; Usb power FET ON
	bcf		INTCON, INTF
	bcf		INTCON, INTE

	movlw	B'00001100'			; PWM ON
	movwf	CCP1CON	
	bsf		T2CON, TMR2ON

	clrf	adstat				; Fix problem, device may not wakeup when operates CC before sleep.
	movlw	B'10001000'			; 32Tosc Clock, Disable A/D, Select channel 1
	movwf	ADCON0
	bcf		T1CON, TMR1ON			; Timer1 off
	bcf		PIR1, TMR1IF			; Clear TMR1IF
	bcf		PIR1, ADIF			; Clear A/D IF
	
	bsf		INTCON, GIE			; Enable INT

check_sleep_end

; ========================================================================================
; A/D conversion
; Most part of A/D conversions are processed in interrupt routine.
; Because A/D takes some time. Save CPU cycle to speed up entire loop for better reguration.
; Overall conversion (Sample V, Convert V, Sample A, Convert A) takes about 100us including
; interrupt latency etc.
 	btfsc	adstat, ADSTAT_SAMPLE_VOLTAGE
	goto	ad_sync
	
	; Prepare for A/D (first time)
	; Start voltage sampling.
	movlw	B'10000001'			; 32T osc clock, channel 0, A/D enabled
	movwf	ADCON0

	; Setup TMR1
	; TMR1 clocked by 5MHz(=0.2us). At least 15us required for accurate sampling.
	; TMR1 INT will be invoked when TMR1 counter overflows. 
	movlw	TSAMPLE_H
	movwf	TMR1H
	movlw	TSAMPLE_L
	movwf	TMR1L
	; Start TMR1
	bsf		T1CON, TMR1ON
	bsf		adstat, ADSTAT_SAMPLE_VOLTAGE

ad_sync
	; Syncronize main loop with A/D conversion.
	btfss	adstat, ADSTAT_DONE
	goto	ad_sync

	; Copy A/D value to current reading register
	movfw	advread_h
	movwf	vread_h
	movfw	advread_l
	movwf	vread_l
	movfw	adaread_h
	movwf	aread_h
	movfw	adaread_l
	movwf	aread_l

	; Prepare for A/D (repeat)
	; Start voltage sampling.
	movlw	B'10000001'			; 32Tosc clock, channel 0, A/D enabled
	movwf	ADCON0

	; Setup TMR1
	movlw	TSAMPLE_H
	movwf	TMR1H
	movlw	TSAMPLE_L
	movwf	TMR1L
	
	; Start TMR1
	bsf		T1CON, TMR1ON

	clrf	adstat
	bsf		adstat, ADSTAT_SAMPLE_VOLTAGE

ad_done

; ========================================================================================
; Power source management
; PORTA:ADPTR goes HIGH, when adaptor is attached, stop USB power feed
; Check VCC presence using PORTB:VCCOK
	btfss	PORTA, ADPTR
	goto	pmng_no_adaptor

	; Adaptor plugged in
	bsf		PORTB, USBPWR			; Turn off USB power FET
	bsf		psustat, F_ADAPTR		; Set adapter detection flag

	; Check VCC
	btfss	PORTB, VCCOK
	goto	pmng_vcc_ng
	
	; VCC OK
	bsf		psustat, F_VCCOK
	goto	pmng_end

pmng_vcc_ng
	; VCC NOT OK
	bcf		psustat, F_VCCOK
	bcf		psustat, F_OUTPUT
	clrf	vread_h				; Clear A/D results
	clrf	vread_l				; A/D shows inproper results, since no VREF present
	clrf	aread_h
	clrf	aread_l
	goto	pmng_end

pmng_no_adaptor
	; No adaptor 
	btfss	psustat, F_ADAPTR
	goto	pmng_no_adaptor_0		

	bcf		psustat, F_OUTPUT
	clrf	vset_h				; Clear setting
	clrf	vset_l				; We can not use same setting when source changed from
	clrf	aset_h				; adaptor to USB power
	clrf	aset_l

pmng_no_adaptor_0
	bcf		PORTB, USBPWR			; Turn on USB power FET
	bcf		psustat, F_ADAPTR
	bsf		psustat, F_VCCOK
	
pmng_end

; ========================================================================================
; Serial communication to PC
	call	comm

; ========================================================================================
; Output sw handler
	btfss	PORTB, OUTSW
	goto	outputsw_pushed

	; outputsw released
	movlw	D'100'				; Set stable counter
	movwf	outputsw_cnt

outputsw_stable
	btfsc	psustat, F_OUTPUT
	goto	outputsw_on

	; outputsw off
	bcf		PORTB, CVLED			; Turn off LEDs
	bcf		PORTB, CCLED
	clrf	vdiff_pre1_h
	clrf	vdiff_pre1_l
	clrf	adiff_pre1_h
	clrf	adiff_pre1_l
	goto	setpwm_zero			; Set pwm to 0

outputsw_pushed
	; outputsw pushed
	movfw	outputsw_cnt
	btfsc	STATUS, Z
	goto	outputsw_stable		; Push continuously

	decfsz	outputsw_cnt, F
	goto	outputsw_stable
	
	movfw	psustat
	xorlw	OUTPUT_MASK			; Reverse it
	movwf	psustat
	goto	outputsw_stable

outputsw_on

; ========================================================================================
; Calc voltage difference (vdiff = vread - vset)
	movfw	vread_h
	movwf	_AH
	movfw	vread_l
	movwf	_AL

	movfw	vset_h
	movwf	_BH
	movfw	vset_l
	movwf	_BL

	call	w_sub
	
	movfw	_AH
	movwf	vdiff_h
	movfw	_AL
	movwf	vdiff_l

; ========================================================================================
; Calc current difference (adiff = aread - aset)
	movfw	aread_h
	movwf	_AH
	movfw	aread_l
	movwf	_AL

	movfw	aset_h
	movwf	_BH
	movfw	aset_l
	movwf	_BL

	call	w_sub
	
	movfw	_AH
	movwf	adiff_h
	movfw	_AL
	movwf	adiff_l

; ========================================================================================
; Determine operating mode (CV/CC)
	btfss	vdiff_h, 7		; If vdiff >= 0 then CV mode 
	goto	cvmode
	
	btfss	adiff_h, 7
	goto	ccmode			; If adiff >= 0 then CC mode

							; vdiff < 0 && adiff < 0
	movfw	vdiff_h
	movwf	_AH
	movfw	vdiff_l
	movwf	_AL
	
	movfw	adiff_h
	movwf	_BH
	movfw	adiff_l
	movwf	_BL
							
	call	w_sub			; -vdiff - -adiff

	btfsc	_AH, 7
	goto	ccmode			; (-vdiff - -adiff) <  0 ... |vdiff| >  |adiff|  => CC mode 
	;goto	cvmode			; (-vdiff - -adiff) >= 0 ... |vdiff| <= |adiff|  => CV mode

; ========================================================================================
; CV mode calc PWM
cvmode
	bcf		psustat, F_CCMODE
	bsf		PORTB, CVLED
	bcf		PORTB, CCLED
	
	clrf	adiff_pre1_h
	clrf	adiff_pre1_l
	
	; PI control
	; dpwm =  2vdiff - vdiff_pre1
	clrf	_AH
	clrf	_AL
	movfw	vdiff_h
	movwf	_BH
	movfw	vdiff_l
	movwf	_BL
	call	w_add
	call	w_add

	movfw	vdiff_pre1_h
	movwf	_BH
	movfw	vdiff_pre1_l
	movwf	_BL
	call	w_sub

	;call	w_mul2
	;call	w_div2

	movfw	_AH
	movwf	dpwm_h
	movfw	_AL
	movwf	dpwm_l

	movfw	vdiff_h
	movwf	vdiff_pre1_h
	movfw	vdiff_l
	movwf	vdiff_pre1_l

	goto	setpwm	

; ========================================================================================
; CC mode calc PWM
ccmode
	bsf		psustat, F_CCMODE
	bsf		PORTB, CCLED
	bcf		PORTB, CVLED
	
	clrf	vdiff_pre1_h
	clrf	vdiff_pre1_l

	; PI control
	; dpwm = (2adiff - adiff_pre1) / coeff
	clrf	_AH
	clrf	_AL
	movfw	adiff_h
	movwf	_BH
	movfw	adiff_l
	movwf	_BL
	call	w_add
	call	w_add

	movfw	adiff_pre1_h
	movwf	_BH
	movfw	adiff_pre1_l
	movwf	_BL
	call	w_sub

	; Change coefficient according to output voltage.
	; if vread > 5.10 then coef = 2^4
	movfw	vread_h
	btfss	STATUS, Z
	goto	ccmode_0

	; if vread > 2.54 then coef = 2^4
	btfsc	vread_l, 7
	goto	ccmode_0

	;if vread <= 2.54 then coef = 2^6
	call	w_div16

ccmode_0
	call	w_div2
	call	w_div2

	movfw	_AH
	movwf	dpwm_h
	movfw	_AL
	movwf	dpwm_l

	movfw	adiff_h
	movwf	adiff_pre1_h
	movfw	adiff_l
	movwf	adiff_pre1_l

; ========================================================================================
; Set PWM
setpwm
	; Get max pwm duty from output voltage.
	; This is important to avoid core magnetic saturation.
	; Max duty can be obtained by following formula.
	; maxduty = Vout / (Vout + Vin * sqrt(L2/L1) )
	; This formula require some CPU power (Using mul,div ...)
	; Use tiny table to save CPU time.
	; Notes: 
	;  Should consider voltage drop (by line ,polysw ...) for Vin. Use 4V instead.
	;  Vout must includes voltage drop caused by diode = 0.5V
	;  Also consider some margin = 1.2
	;  L1 = L2 = 2.6E-5, MaxPWM = 1024 * maxduty
	;  Thus, MaxPWM = 1024 * 1.2 * (Vout + 0.5) / ((Vout + 0.5) + 4)
	;  Maxduty shoud be limited to 0.7 (Pin=7.57W at 4V / B=2850Gauss at 5V < CoreMax B=3800Gauss at 100C)
	movfw	vread_h
	btfsc	STATUS, Z
	goto	setpwm_max_low
	
	andlw	0x01
	btfss	STATUS, Z
	goto	setpwm_max_mid
	goto	setpwm_max_high

setpwm_max_low
	; 0 <= vread < 256 (0V <= vout < 5.12 V)
	call	vread2maxpwm_low_l
	movwf	maxpwm_l
	call	vread2maxpwm_low_h
	movwf	maxpwm_h
	goto	setpwm_max_end

setpwm_max_mid
	; 256 <= vread < 512 (5.12V <= vout < 10.24V)
	call	vread2maxpwm_mid_l
	movwf	maxpwm_l
	call	vread2maxpwm_mid_h
	movwf	maxpwm_h
	goto	setpwm_max_end

setpwm_max_high
	; 512 <= vread < 1024 (10.24V <= vout < 20.48V) 
	call	vread2maxpwm_high_l
	movwf	maxpwm_l
	call	vread2maxpwm_high_h
	movwf	maxpwm_h
	
setpwm_max_end
	; Current PWM value => _A
	; CCPR1L holds Upper 8bit and Lower 2bit at CCP1CON 5:4
	swapf	CCP1CON, W
	andlw	B'00000011'	
	movwf	_AL
	
	rlf		CCPR1L, W
	movwf	tmp_l
	rlf		tmp_l, W
	andlw	B'11111100'
	addwf	_AL, F

	swapf	CCPR1L, W
	movwf	tmp_l
	rrf		tmp_l, F
	rrf		tmp_l, W
	andlw	B'00000011'
	movwf	_AH

	; Delta PWM value => _B
	movfw	dpwm_h
	movwf	_BH
	movfw	dpwm_l
	movwf	_BL

	; New PWM value (_A) = Current PWM - Delta PWM
	call	w_sub

	; _B = _A
	movfw	_AH
	movwf	_BH
	movfw	_AL
	movwf	_BL

	; Check New PWM range
	btfsc	_BH, 7
	goto	setpwm_zero		; Can not set negative use 0 instead

	; _A = Max PWM
	movfw	maxpwm_h
	movwf	_AH
	movfw	maxpwm_l
	movwf	_AL

	; _A = Max PWM - New PWM
	call	w_sub

	btfsc	_AH, 7
	goto	setpwm_max		; Limit with Max PWM
	goto	setpwm_reg		; In range

setpwm_zero
	clrf	_BH
	clrf	_BL
	goto	setpwm_reg

setpwm_max
	movfw	maxpwm_h
	movwf	_BH
	movfw	maxpwm_l
	movwf	_BL

setpwm_reg
	; Set New PWM
	; Upper 8 bit	
	swapf	_BH, F
	rlf		_BH, F
	rlf		_BH, W
	andlw	B'11000000'
	movwf	_BH
	
	rrf		_BL, W
	movwf	tmp_l
	rrf		tmp_l, W
	andlw	B'00111111'

	addwf	_BH, W
	movwf	CCPR1L

	; Lower 2 bit
	swapf	_BL, W
	andlw	B'00110000'
	movwf	tmp_l

	movfw	CCP1CON
	andlw	B'00001111'
	addwf	tmp_l, W
	movwf	CCP1CON

; ========================================================================================
; Return to loop head
	goto	loop

; ========================================================================================
; USBPSU main interrupt handler
; Input reg    : 
; Output reg   : 
; Internal reg : 
; Remarks      : 
; ========================================================================================
isr_main
	; Other interrupts will be caused by TMR1 or A/D
	; Don't use A/D reson flag, because we know what causes in advance.   
	btfsc	adstat, ADSTAT_CONVERT_CURRENT
	goto	isr_main_convert_current
	btfsc	adstat, ADSTAT_SAMPLE_CURRENT
	goto	isr_main_sample_current
	btfsc	adstat, ADSTAT_CONVERT_VOLTAGE
	goto	isr_main_convert_voltage

isr_main_sample_voltage
	; Sampling (voltage) done.
	; Prepare for A/D conversion
	bsf		ADCON0, GO		; Start A/D
	bcf		T1CON, TMR1ON		; Timer1 off
	bcf		PIR1, TMR1IF		; Clear TMR1 INT flag
	;bcf		PIR1, ADIF		; Clear A/D INT flag
	bsf		adstat, ADSTAT_CONVERT_VOLTAGE	; Set next state
	;bsf		ADCON0, GO		; Start A/D
	goto	RESTORE
	
isr_main_convert_voltage
	; A/D conversion (voltage) done.
	; Start current sampling.
	movlw	B'10001001'		; 32Tosc Clock, Enable A/D, Select channel 1
	movwf	ADCON0

	; Setup TMR1
	; TMR1 INT will be invoked when TMR1 counter overflows. 
	movlw	TSAMPLE_H
	movwf	TMR1H
	movlw	TSAMPLE_L
	movwf	TMR1L
	
	; Start TMR1
	bsf		T1CON, TMR1ON

	;bcf		PIR1, TMR1IF		; Clear TMR1 INT flag
	bcf		PIR1, ADIF		; Clear A/D INT flag

	; Save A/D result
	banksel	ADRESL
	movfw	ADRESL
	banksel	advread_l
	movwf	advread_l
	movfw	ADRESH
	movwf	advread_h

	bsf		adstat, ADSTAT_SAMPLE_CURRENT	; Set next state
	goto	RESTORE

isr_main_sample_current
	; Sampling (current) done.
	bsf		ADCON0, GO		; Start A/D
	bcf		T1CON, TMR1ON		; Timer1 off
	bcf		PIR1, TMR1IF		; Clear TMR1 INT flag
	;bcf		PIR1, ADIF		; Clear A/D INT flag
	bsf		adstat, ADSTAT_CONVERT_CURRENT	; Set next state
	;bsf		ADCON0, GO		; Start A/D
	goto	RESTORE

isr_main_convert_current
	; A/D conversion (current) done.

	;bcf		PIR1, TMR1IF		; Clear TMR1 INT flag
	bcf		PIR1, ADIF		; Clear A/D INT flag

	banksel	ADRESL
	movfw	ADRESL
	banksel	adaread_l
	movwf	adaread_l
	movfw	ADRESH
	movwf	adaread_h
	bsf		adstat, ADSTAT_DONE
	goto	RESTORE

; ========================================================================================
; 16bit add (_A = _A + _B)
; Input reg    : _AH, _AL, _BH, _BL
; Output reg   : _AH, _AL
; Internal reg : None
; Remarks      : See http://ha6.seikyou.ne.jp/home/sasao/junkhard/pic/#012
; ========================================================================================
w_add
	movf	_BL,W	;  8biťvZ
	addwf	_AL,F	; _AL = _AL + _BL

	movf	_BH,W	;  8biťvZ
	btfsc	STATUS,C
	incf	_BH,W	; _AH = _AH + _BH (+ 1)
	addwf	_AH,F	; * (+1)͌Jオ莞̂
	return

; ========================================================================================
; 16bit sub (_A = _A - _B)
; Input reg    : _AH, _AL, _BH, _BL
; Output reg   : _AH, _AL
; Internal reg : None
; Remarks      : See http://ha6.seikyou.ne.jp/home/sasao/junkhard/pic/#012
; ========================================================================================
w_sub
	movf	_BL,W	;  8biťvZ
	subwf	_AL,F	; _AL = _AL - _BL

	movf	_BH,W	;  8biťvZ
	btfss	STATUS,C
	incf	_BH,W	; _AH = _AH - {_BH (+1)}
	subwf	_AH,F	; * (+1)͌؂莞̂
	return

w_mul2
	bcf		STATUS, C
	rlf		_AL, F
	rlf		_AH, F
	return

; ========================================================================================
; 16bit divide by 2 (_A = _A / 2)
; Input reg    : _AH, _AL
; Output reg   : _AH, _AL
; Internal reg : None
; Remarks      : Return 1 if input is 1
; ========================================================================================
w_div2
	btfsc	_AH, 7
	goto	w_div2_neg

	movf	_AH, F
	btfss	STATUS, Z
	goto	w_div_0

	movf	_AL, F
	btfsc	STATUS, Z
	return	; _A = 0

	; _AH=0, _AL>0
	bcf		STATUS, C
	rrf		_AL, F
	movf	_AL, F
	btfss	STATUS, Z
	return

	incf	_AL, F
	return

w_div_0
	bcf		STATUS, C
	rrf		_AH, F
	rrf		_AL, F
	return
	
w_div2_neg
	bsf		STATUS, C
	rrf		_AH, F
	rrf		_AL, F
	return

; ========================================================================================
; 16bit divide by 16 (_A = _A / 16)
; Input reg    : _AH, _AL
; Output reg   : _AH, _AL
; Internal reg : None
; Remarks      : Return 1 if input is under 16
; ========================================================================================
w_div16
	btfsc	_AH, 7
	goto	w_div16_neg

	movf	_AH, F
	btfsc	STATUS, Z
	goto	w_div16_0

	swapf	_AL, W
	andlw	0x0f
	movwf	_AL

	swapf	_AH, W
	andlw	0xf0
	addwf	_AL, F
	
	clrf	_AH
	return
	
w_div16_0
	movfw	_AL
	btfsc	STATUS, Z
	return

	movfw	_AL
	andlw	0xf0
	btfsc	STATUS, Z
	goto	w_div16_1

	swapf	_AL, W
	andlw	0x0f
	movwf	_AL
	return

w_div16_1
	clrf	_AL
	incf	_AL, F
	return

w_div16_neg
	swapf	_AL, W
	andlw	0x0f
	movwf	_AL

	swapf	_AH, W
	andlw	0xf0
	addwf	_AL, F

	movlw	0xff
	movwf	_AH
	return

; ========================================================================================
; Serial command controller
; Input reg    : vset_h, vset_l, aset_h, aset_l, vread_h, vread_l, aread_h, aread_l, psustat
; Output reg   : TXREG, RXREG, vset_h, vset_l, aset_h, aset_l, psustat
; Internal reg : txleft, rxleft, comstat, _AH, _AL, _BH, _BL, FSR, INDF
;              : buff_0 - buff_10
; Remarks      : 
; ========================================================================================
comm	
	btfsc	TXBUSY				; Do nothind if TX is in progress
	return

	movf	txleft, F				; Do tx if there are data to be sent
	btfsc	STATUS, Z
	goto	comm_rx	

	movfw	INDF
	movwf	TXREG

	incf	FSR, F
	decf	txleft, F

	btfsc	STATUS, Z				; Prepare RX, if no more data
	bcf		comstat, RXWAITING 

	call	PUTCHAR
	return

comm_rx
	; Recieve 
	btfss	comstat, RXWAITING		; if (!RXWAITING) call GETCHAR
	call	GETCHAR
	bsf		comstat, RXWAITING

	btfsc	RXERR				; if (RXERR) goto comm_rxerr
	goto	comm_rxerr

	btfss	RECV					; Check data arrival
	return						; No data

	movf	rxleft, F				; if (rxleft != 0) goto comm_getoperand
	btfss	STATUS, Z
	goto	comm_getoperand

	; Command dispatcher
	movlw	"M"					; M => Monitor current status
	xorwf	RXREG, W
	btfsc	STATUS, Z
	goto	comm_getstatus

	movlw	"S"					; S => Set Voltage, Current
	xorwf	RXREG, W
	btfsc	STATUS, Z
	goto	comm_setparam	

	movlw	"N"					; N => Output ON
	xorwf	RXREG, W
	btfsc	STATUS, Z
	goto	comm_out_on
	
	movlw	"F"					; F => Output OFF
	xorwf	RXREG, W
	btfsc	STATUS, Z
	goto	comm_out_off

	movlw	"V"					; V => Firmware version
	xorwf	RXREG, W
	btfsc	STATUS, Z
	goto	comm_getversion

	movlw	"I"					; I => Device identification
	xorwf	RXREG, W
	btfsc	STATUS, Z
	goto	comm_ident

	movlw	"L"					; L = > Get limitation
	xorwf	RXREG, W
	btfsc	STATUS, Z
	goto	comm_getlimit
	
	movlw	"D"					; D => Debug use
	xorwf	RXREG, W
	btfsc	STATUS, Z
	goto	comm_debug

	goto	comm_ack_ng			; Others
	
comm_rxerr
	call	GETCHAR
	return

comm_getoperand
	; Command S has 6 bytes operand.
	; Where leading 3 bytes represents voltage, and left 3bytes represents current.
	; Operand are stored in buffer (0x70 to 0x75)
	movfw	RXREG
	movwf	INDF

	incf	FSR, F
	decf	rxleft, F				; rxleft --

	btfss	STATUS, Z				; if (rxleft > 0) goto comm_getoperand_more
	goto	comm_getoperand_more

	; Apply voltage
	movfw	buff_0					; if (Recieved voltage > MAXVSET) goto comm_ack_ng
	call	hex2int
	movwf	_BH
	
	movfw	buff_1
	call	hex2int
	movwf	_BL
	swapf	_BL, F

	movfw	buff_2
	call	hex2int
	addwf	_BL, F

	; Check max limit
	movlw	MAXVSET_H
	movwf	_AH
	movlw	MAXVSET_L
	movwf	_AL

	call	w_sub
	btfsc	_AH, 7
	goto	comm_ack_ng

	; OK. Set vset register
	movfw	_BH
	movwf	vset_h
	movfw	_BL
	movwf	vset_l

	; Apply current
	movfw	buff_3					; if (Recieved current > MAXASET) goto comm_ack_ng
	call	hex2int
	movwf	_BH
	
	movfw	buff_4
	call	hex2int
	movwf	_BL
	swapf	_BL, F

	movfw	buff_5
	call	hex2int
	addwf	_BL, F

	; Check max limit
	movlw	MAXASET_H
	movwf	_AH
	movlw	MAXASET_L
	movwf	_AL

	call	w_sub
	btfsc	_AH, 7
	goto	comm_ack_ng

	; OK. Set aset register
	movfw	_BH
	movwf	aset_h
	movfw	_BL
	movwf	aset_l
	
	; OK
	goto	comm_ack_ok

comm_getoperand_more
	call	GETCHAR
	return

comm_setparam
	; Set Voltage and Current
	; This command has 6 bytes operand
	; Prepare to recieve operand
	movlw	0x06			; Length of operand
	movwf	rxleft
	movlw	buff_0			; Set buffer pointer
	movwf	FSR
	call	GETCHAR
	return

comm_getstatus
	; Get current status
	; PSMVVVAAA<CRLF>, length=11
	;  P: U=USB Power / A=Adapter Power / N=Adapter present but not powered
	;  S: N=output ON /F=output OFF
 	;  M: V=mode CV / C=mode CC
 	; VVV: Voltage reading (000`3FF)
 	; AAA: Current reading (000`3FF)

	; Power source
	movlw	"U"
	btfss	psustat, F_ADAPTR
	goto	com_getstatus_0
	; Adapter power
	btfss	psustat, F_VCCOK
	movlw	"N"	
	btfsc	psustat, F_VCCOK
	movlw	"A"
com_getstatus_0
	movwf	buff_0

	; Output SW
	btfss	psustat, F_OUTPUT
	movlw	"F"
	btfsc	psustat, F_OUTPUT
	movlw	"N"
	movwf	buff_1

	; Mode
	btfss	psustat, F_CCMODE
	movlw	"V"
	btfsc	psustat, F_CCMODE
	movlw	"C"
	movwf	buff_2

	; Voltage
	movfw	vread_h
	call	int2hex
	movwf	buff_3

	swapf	vread_l, W
	call	int2hex
	movwf	buff_4

	movfw	vread_l
	call	int2hex
	movwf	buff_5

	; Current
	movfw	aread_h
	call	int2hex
	movwf	buff_6

	swapf	aread_l, W
	call	int2hex
	movwf	buff_7

	movfw	aread_l
	call	int2hex
	movwf	buff_8

	; <CRLF>
	movlw	0x0d
	movwf	buff_9
	movlw	0x0a
	movwf	buff_10
	
	movlw	0x0b
	movwf	txleft
	movlw	buff_0
	movwf	FSR
	return

comm_getlimit
	; Get current limit
	; VVVAAA<CRLF>, length=8
 	; VVV: Limit value (000`3FF)
 	; AAA: Limit value (000`3FF)

	; Voltage
	movfw	vset_h
	call	int2hex
	movwf	buff_0

	swapf	vset_l, W
	call	int2hex
	movwf	buff_1

	movfw	vset_l
	call	int2hex
	movwf	buff_2

	; Current
	movfw	aset_h
	call	int2hex
	movwf	buff_3

	swapf	aset_l, W
	call	int2hex
	movwf	buff_4

	movfw	aset_l
	call	int2hex
	movwf	buff_5

	; <CRLF>
	movlw	0x0d
	movwf	buff_6
	movlw	0x0a
	movwf	buff_7
	
	movlw	0x08
	movwf	txleft
	movlw	buff_0
	movwf	FSR
	return

comm_out_on
	; Output ON
	bsf		psustat, F_OUTPUT
	goto	comm_ack_ok

comm_out_off
	; Output OFF
	bcf		psustat, F_OUTPUT
	goto	comm_ack_ok

comm_getversion
	; Send Version
	; X.XX<CRLF>, length=6
	movlw	VERSION_L1
	movwf	buff_0
	movlw	VERSION_L2
	movwf	buff_1
	movlw	VERSION_L3
	movwf	buff_2
	movlw	VERSION_L4
	movwf	buff_3

	; <CRLF>
	movlw	0x0d
	movwf	buff_4
	movlw	0x0a
	movwf	buff_5

	movlw	0x06
	movwf	txleft
	movlw	buff_0
	movwf	FSR
	return

comm_ack_ok
	; Send OK
	movlw	"O"
	movwf	buff_0
	movlw	"K"
	movwf	buff_1

	; <CRLF>
	movlw	0x0d
	movwf	buff_2
	movlw	0x0a
	movwf	buff_3
	
	movlw	4
	movwf	txleft
	movlw	buff_0
	movwf	FSR
	return

comm_ack_ng
	; Send NG
	movlw	"N"
	movwf	buff_0
	movlw	"G"
	movwf	buff_1

	; <CRLF>
	movlw	0x0d
	movwf	buff_2
	movlw	0x0a
	movwf	buff_3
	
	movlw	4
	movwf	txleft
	movlw	buff_0
	movwf	FSR
	return	

comm_ident
	; Send USBPSU
	movlw	"U"
	movwf	buff_0
	movlw	"S"
	movwf	buff_1
	movlw	"B"
	movwf	buff_2
	movlw	"P"
	movwf	buff_3
	movlw	"S"
	movwf	buff_4
	movlw	"U"
	movwf	buff_5

	; <CRLF>
	movlw	0x0d
	movwf	buff_6
	movlw	0x0a
	movwf	buff_7

	movlw	8
	movwf	txleft

	movlw	buff_0
	movwf	FSR
	return

comm_debug
	; Debug purpose
	; Get difference (for debug)
	; VVVVAAAA<CRLF>

	; Voltage
	swapf	dpwm_h, W
	andlw	0x0f
	call	int2hex
	movwf	buff_0

	movfw	dpwm_h
	andlw	0x0f
	call	int2hex
	movwf	buff_1

	swapf	dpwm_l, W
	andlw	0x0f
	call	int2hex
	movwf	buff_2

	movfw	dpwm_l
	andlw	0x0f
	call	int2hex
	movwf	buff_3

	; Current
	swapf	dbg_h, W
	andlw	0x0f
	call	int2hex
	movwf	buff_4

	movfw	dbg_h
	andlw	0x0f
	call	int2hex
	movwf	buff_5

	swapf	dbg_l, W
	andlw	0x0f
	call	int2hex
	movwf	buff_6

	movfw	dbg_l
	andlw	0x0f
	call	int2hex
	movwf	buff_7

	; <CRLF>
	movlw	0x0d
	movwf	buff_8
	movlw	0x0a
	movwf	buff_9
	
	movlw	0x0a
	movwf	txleft
	movlw	buff_0
	movwf	FSR
	return

; ========================================================================================
; Table resides over 0x700 to avoid PCL boundary problem
; ========================================================================================
	ORG		0x700

; ========================================================================================
; Convert hex to character (w <= toChar(w))
; Input reg    : w
; Output reg   : w
; Internal reg : 
; Remarks      : 
; ========================================================================================
int2hex
	;movwf	tmp_l
	;movlw	0x07
	;movwf	PCLATH
	;movfw	tmp_l
	andlw	0x0f
	addwf	PCL, F
	retlw	"0"
	retlw	"1"
	retlw	"2"
	retlw	"3"
	retlw	"4"
	retlw	"5"
	retlw	"6"
	retlw	"7"
	retlw	"8"
	retlw	"9"
	retlw	"A"
	retlw	"B"
	retlw	"C"
	retlw	"D"
	retlw	"E"
	retlw	"F"

; ========================================================================================
; Get max PWM(H) duty from output voltage (0 <= vread < 256 (0V <= vout < 5.12 V))
; Input reg    : vread_l 
; Output reg   : w
; Internal reg : 
; Remarks      : 
; ========================================================================================
vread2maxpwm_low_h
	;movlw	0x07
	;movwf	PCLATH
	swapf	vread_l, W
	andlw	0x0f
	addwf	PCL, F
	retlw	D'0'		; Vo <= 0.30 
	retlw	D'0'		; Vo <= 0.62
	retlw	D'1'		; Vo <= 0.94
	retlw	D'1'		; Vo <= 1.26		
	retlw	D'1'		; Vo <= 1.58
	retlw	D'1'		; Vo <= 1.90
	retlw	D'1'		; Vo <= 2.22
	retlw	D'1'		; Vo <= 2.54
	retlw	D'1'		; Vo <= 2.86
	retlw	D'1'		; Vo <= 3.18
	retlw	D'2'		; Vo <= 3.50
	retlw	D'2'		; Vo <= 3.82
	retlw	D'2'		; Vo <= 4.14
	retlw	D'2'		; Vo <= 4.46
	retlw	D'2'		; Vo <= 4.78
	retlw	D'2'		; Vo <= 5.10

; ========================================================================================
; Get max PWM(L) duty from output voltage (0 <= vread < 256 (0V <= vout < 5.12 V))
; Input reg    : vread_l 
; Output reg   : w
; Internal reg : 
; Remarks      : 
; ========================================================================================
vread2maxpwm_low_l
	;movlw	0x07
	;movwf	PCLATH
	swapf	vread_l, W
	andlw	0x0f
	addwf	PCL, F
	retlw	D'171'	; Vo <= 0.30
	retlw	D'224'	; Vo <= 0.62
	retlw	D'015'	; Vo <= 0.94
	retlw	D'057'	; Vo <= 1.26
	retlw	D'094'	; Vo <= 1.58
	retlw	D'128'	; Vo <= 1.90
	retlw	D'158'	; Vo <= 2.22
	retlw	D'186'	; Vo <= 2.54
	retlw	D'211'	; Vo <= 2.86
	retlw	D'235'	; Vo <= 3.18
	retlw	D'000'	; Vo <= 3.50
	retlw	D'020'	; Vo <= 3.82
	retlw	D'038'	; Vo <= 4.14
	retlw	D'055'	; Vo <= 4.46
	retlw	D'071'	; Vo <= 4.78
	retlw	D'085'	; Vo <= 5.10

; ========================================================================================
; Get max PWM(H) duty from output voltage (256 <= vread < 512 (5.12V <= vout < 10.24V))
; Input reg    : vread_l 
; Output reg   : w
; Internal reg : tmp_l
; Remarks      : 
; ========================================================================================
vread2maxpwm_mid_h
	retlw	D'002'

; ========================================================================================
; Get max PWM(L) duty from output voltage (256 <= vread < 512 (5.12V <= vout < 10.24V))
; Input reg    :  
; Output reg   : w
; Internal reg : 
; Remarks      : 
; ========================================================================================
vread2maxpwm_mid_l
	;movlw	0x07
	;movwf	PCLATH
	swapf	vread_l, W
	andlw	0x0f
	addwf	PCL, F
	retlw	D'099'	; Vo <= 5.42
	retlw	D'112'	; Vo <= 5.74
	retlw	D'124'	; Vo <= 6.06
	retlw	D'136'	; Vo <= 6.38
	retlw	D'146'	; Vo <= 6.70
	retlw	D'154'	; Vo <= 7.02
	retlw	D'154'	; Vo <= 7.34
	retlw	D'154'	; Vo <= 7.66
	retlw	D'154'	; Vo <= 7.98
	retlw	D'154'	; Vo <= 8.30
	retlw	D'154'	; Vo <= 8.62
	retlw	D'154'	; Vo <= 8.94
	retlw	D'154'	; Vo <= 9.26
	retlw	D'154'	; Vo <= 9.58
	retlw	D'154'	; Vo <= 9.90
	retlw	D'154'	; Vo <= 10.22

; ========================================================================================
; Get max PWM(H) duty from output voltage (512 <= vread < 1024 (10.24V <= vout < 20.48V))
; Input reg    :  
; Output reg   : w
; Internal reg : 
; Remarks      : 
; ========================================================================================
vread2maxpwm_high_h
	retlw	D'002'

; ========================================================================================
; Get max PWM(L) duty from output voltage (512 <= vread < 1024 (10.24V <= vout < 20.48V))
; Input reg    : vread_l 
; Output reg   : w
; Internal reg : 
; Remarks      : 
; ========================================================================================
vread2maxpwm_high_l
	retlw	D'154'

; ---------------------------------------------------------
; EEPROM Default data
; ---------------------------------------------------------
	ORG	0x2100

; ---------------------------------------------------------
; Directive 'end of program'
; ---------------------------------------------------------
	END
